@@ -1,4 +1,5 @@
 # -*- encoding: utf-8 -*-
+from collections import OrderedDict
 from io import BytesIO
 
 import mock
@@ -72,6 +73,50 @@ class HTTPClientTests(TestCase):
             None,
         )
 
+    def test_request_uri_bytes_pass(self):
+        """
+        The URL parameter may contain path segments or querystring parameters
+        that are not valid UTF-8. These pass through.
+        """
+        # This URL is http://example.com/hello?who=you, but "hello", "who", and
+        # "you" are encoded as UTF-16. The particulars of the encoding aren't
+        # important; what matters is that those segments can't be decoded by
+        # Hyperlink's UTF-8 default.
+        self.client.request(
+            "GET",
+            (
+                "http://example.com/%FF%FEh%00e%00l%00l%00o%00"
+                "?%FF%FEw%00h%00o%00=%FF%FEy%00o%00u%00"
+            ),
+        )
+        self.agent.request.assert_called_once_with(
+            b'GET',
+            (
+                b'http://example.com/%FF%FEh%00e%00l%00l%00o%00'
+                b'?%FF%FEw%00h%00o%00=%FF%FEy%00o%00u%00'
+            ),
+            Headers({b'accept-encoding': [b'gzip']}),
+            None,
+        )
+
+    def test_request_uri_plus_pass(self):
+        """
+        URL parameters may contain spaces encoded as ``+``. These remain as
+        such and are not mangled.
+
+        This reproduces `Klein #339 <https://github.com/twisted/klein/issues/339>`_.
+        """
+        self.client.request(
+            "GET",
+            "https://example.com/?foo+bar=baz+biff",
+        )
+        self.agent.request.assert_called_once_with(
+            b'GET',
+            b"https://example.com/?foo+bar=baz+biff",
+            Headers({b'accept-encoding': [b'gzip']}),
+            None,
+        )
+
     def test_request_uri_idn_params(self):
         """
         A URL that contains non-ASCII characters can be augmented with
@@ -159,9 +204,13 @@ class HTTPClientTests(TestCase):
         treq coerces non-string param names passed to *params* like
         `urllib.urlencode()`
         """
+        # A value used to test that it is never encoded or decoded.
+        # It should be invalid UTF-8 or UTF-32 (at least).
+        raw_bytes = b"\x00\xff\xfb"
+
         self.client.request('GET', 'http://example.com/', params=[
             (u'text', u'A\u03a9'),
-            (b'bytes', ['ascii']),
+            (b'bytes', ['ascii', raw_bytes]),
             ('native', 'native'),
             (1, 'int'),
             (None, ['none']),
@@ -171,7 +220,7 @@ class HTTPClientTests(TestCase):
             b'GET',
             (
                 b'http://example.com/'
-                b'?text=A%CE%A9&bytes=ascii'
+                b'?text=A%CE%A9&bytes=ascii&bytes=%00%FF%FB'
                 b'&native=native&1=int&None=none'
             ),
             Headers({b'accept-encoding': [b'gzip']}),
@@ -397,6 +446,36 @@ class HTTPClientTests(TestCase):
                 boundary=b'heyDavid'),
             self.MultiPartProducer.call_args)
 
+    def test_request_files_tuple_too_short(self):
+        """
+        The `HTTPClient.request()` *files* argument requires tuples of length
+        2 or 3. It raises `TypeError` when the tuple is too short.
+        """
+        with self.assertRaises(TypeError) as c:
+            self.client.request(
+                "POST",
+                b"http://example.com/",
+                files=[("t1", ("foo.txt",))],
+            )
+
+        self.assertIn("'t1' tuple has length 1", str(c.exception))
+
+    def test_request_files_tuple_too_long(self):
+        """
+        The `HTTPClient.request()` *files* argument requires tuples of length
+        2 or 3. It raises `TypeError` when the tuple is too long.
+        """
+        with self.assertRaises(TypeError) as c:
+            self.client.request(
+                "POST",
+                b"http://example.com/",
+                files=[
+                    ("t4", ("foo.txt", "text/plain", BytesIO(b"...\n"), "extra!")),
+                ],
+            )
+
+        self.assertIn("'t4' tuple has length 4", str(c.exception))
+
     @mock.patch('treq.client.uuid.uuid4', mock.Mock(return_value="heyDavid"))
     def test_request_mixed_params(self):
 
@@ -460,6 +539,50 @@ class HTTPClientTests(TestCase):
                           data=BytesIO(b"yo"),
                           files={"file1": BytesIO(b"hey")})
 
+    def test_request_json_with_data(self):
+        """
+        Passing `HTTPClient.request()` both *data* and *json* parameters is
+        invalid because *json* is ignored. This behavior is deprecated.
+        """
+        self.client.request(
+            "POST",
+            "http://example.com/",
+            data=BytesIO(b"..."),
+            json=None,  # NB: None is a valid value. It encodes to b'null'.
+        )
+
+        [w] = self.flushWarnings([self.test_request_json_with_data])
+        self.assertEqual(DeprecationWarning, w["category"])
+        self.assertEqual(
+            (
+                "Argument 'json' will be ignored because 'data' was also passed."
+                " This will raise TypeError in the next treq release."
+            ),
+            w['message'],
+        )
+
+    def test_request_json_with_files(self):
+        """
+        Passing `HTTPClient.request()` both *files* and *json* parameters is
+        invalid because *json* is ignored. This behavior is deprecated.
+        """
+        self.client.request(
+            "POST",
+            "http://example.com/",
+            files={"f1": ("foo.txt", "text/plain", BytesIO(b"...\n"))},
+            json=["this is ignored"],
+        )
+
+        [w] = self.flushWarnings([self.test_request_json_with_files])
+        self.assertEqual(DeprecationWarning, w["category"])
+        self.assertEqual(
+            (
+                "Argument 'json' will be ignored because 'files' was also passed."
+                " This will raise TypeError in the next treq release."
+            ),
+            w['message'],
+        )
+
     def test_request_dict_headers(self):
         self.client.request('GET', 'http://example.com/', headers={
             'User-Agent': 'treq/0.1dev',
@@ -473,6 +596,81 @@ class HTTPClientTests(TestCase):
                      b'Accept': [b'application/json', b'text/plain']}),
             None)
 
+    def test_request_headers_object(self):
+        """
+        The *headers* parameter accepts a `twisted.web.http_headers.Headers`
+        instance.
+        """
+        self.client.request(
+            "GET",
+            "https://example.com",
+            headers=Headers({"X-Foo": ["bar"]}),
+        )
+
+        self.agent.request.assert_called_once_with(
+            b"GET",
+            b"https://example.com",
+            Headers({
+                "X-Foo": ["bar"],
+                "Accept-Encoding": ["gzip"],
+            }),
+            None,
+        )
+
+    def test_request_headers_invalid_type(self):
+        """
+        `HTTPClient.request()` warns that headers of an unexpected type are
+        invalid and that this behavior is deprecated.
+        """
+        self.client.request('GET', 'http://example.com', headers=[])
+
+        [w] = self.flushWarnings([self.test_request_headers_invalid_type])
+        self.assertEqual(DeprecationWarning, w['category'])
+        self.assertIn(
+            "headers must be a dict, twisted.web.http_headers.Headers, or None,",
+            w['message'],
+        )
+
+    def test_request_dict_headers_invalid_values(self):
+        """
+        `HTTPClient.request()` warns that non-string header values are dropped
+        and that this behavior is deprecated.
+        """
+        self.client.request('GET', 'http://example.com', headers=OrderedDict([
+            ('none', None),
+            ('one', 1),
+            ('ok', 'string'),
+        ]))
+
+        [w1, w2] = self.flushWarnings([self.test_request_dict_headers_invalid_values])
+        self.assertEqual(DeprecationWarning, w1['category'])
+        self.assertEqual(DeprecationWarning, w2['category'])
+        self.assertIn(
+            "The value of headers key 'none' has non-string type",
+            w1['message'],
+        )
+        self.assertIn(
+            "The value of headers key 'one' has non-string type",
+            w2['message'],
+        )
+
+    def test_request_invalid_param(self):
+        """
+        `HTTPClient.request()` warns that invalid parameters are ignored and
+        that this is deprecated.
+        """
+        self.client.request('GET', b'http://example.com', invalid=True)
+
+        [w] = self.flushWarnings([self.test_request_invalid_param])
+        self.assertEqual(
+            (
+                "Got unexpected keyword argument: 'invalid'."
+                " treq will ignore this argument,"
+                " but will raise TypeError in the next treq release."
+            ),
+            w['message'],
+        )
+
     @with_clock
     def test_request_timeout_fired(self, clock):
         """
